module 'data.char' {
   /** check(X).
    *  Succeeds if <code>X</code> is instantiated to a character.
    */

   check(X) :-
      'data.atom':check(X),
      'data.atom':length(X, 1).

   /** code(Char, Code).
    * <prerequisite>
    * <code>Char</code> is an instantiated character.
    * </prerequisite>
    * Succeeds if the integer <code>Code</code> is the code-point
    * of the character <code>Char</code>.
    */

   code(Char, Code) :-
      'kernel':char_to_code(Char, Code).

   /** decimal_digit(Char, Value).
    * Succeeds if <code>Char</code> is instantiated to the character
    * corresponding to the decimal digit
    * (a number between <code>0</code> and <code>9</code>),
    * the value of which is <code>Value</code>
    */

   decimal_digit(Char, Value) :-
      'data.atom':check(Char),
      (
         'data.term':(Char >= '0'),
         'data.term':(Char =< '9')
      ->
         Value = 'data.number':('data.char':code(Char) - 0'0)
      ).

   /** octal_digit(Char, Value).
    * Succeeds if <code>Char</code> is instantiated to the character
    * corresponding to the octal digit (a number between <code>0</code>
    * and <code>7</code>),
    * the value of which is <code>Value</code>
    */

   octal_digit(Char, Value) :-
      'data.atom':check(Char),
      (
         'data.term':(Char >= '0'),
         'data.term':(Char =< '7')
      ->
         Value = 'data.number':('data.char':code(Char) - 0'0)
      ).

   /** hexadecimal_digit(Char, Value).
    * Succeeds if <code>Char</code> is instantiated to the character
    * corresponding to the hexadecimal digit (a number between
    * <code>0</code> and <code>9</code>
    * or <code>a</code> and <code>f</code>, lower or upper case, for decimal
    * values between 0 and 15 respectively),
    * the value of which is <code>Value</code>
    */

   hexadecimal_digit(Char, Value) :-
      'data.atom':check(Char),
      (
         'data.term':(Char >= '0'),
         'data.term':(Char =< '9')
      ->
         Value = 'data.number':('data.char':code(Char) - 0'0)
      ;
         'data.term':(Char >= 'a'),
         'data.term':(Char =< 'f')
      ->
         Value = 'data.number':('data.char':code(Char) - 0'a + 0xa)
      ;
         'data.term':(Char >= 'A'),
         'data.term':(Char =< 'F')
      ->
         Value = 'data.number':('data.char':code(Char) - 0'A + 0xA)
      ).

   /** bit(Char, Value).
    * Succeeds if <code>Char</code> is instantiated to the character
    * corresponding to bit (either <code>0</code> or <code>1</code>),
    * the value of which is <code>Value</code>
    */

   bit(Char, Value) :-
      'data.atom':check(Char),
      (
         'data.term':(Char >= '0'),
         'data.term':(Char =< '1')
      ->
         Value = 'data.number':('data.char':code(Char) - 0'0)
      ).

   /** escape(Char, Escaped).
    *  Succeeds if <code>Escaped</code> is an atom corresponding
    *  to the escaped form of <code>Char</code>.
    */

   escape('\n', '\\n') :-
      !.
   escape('\r', '\\r') :-
      !.
   escape('\t', '\\t') :-
      !.
   escape('\b', '\\b') :-
      !.
   escape('\\', '\\\\') :-
      !.
   escape('\'', '\\\'') :-
      !.
   escape('\"', '\\\"') :-
      !.
   escape(C, C). 

   /** read_escape(Stream, Char).
    *  <prerequisite>
    *  <code>Stream</code> is a stream.
    *  </prerequisite>
    *  Reads an escaped form of a character (excluding the initial
    *  slash) and succeeds if
    *  the corresponding character is <code>Char</code>.
    */

   read_escape(Stream, C) :-
      Stream:read(H),
      (
         H = 'n'
      ->
         C = '\n'
      ;
         H = 'r'
      ->
         C = '\r'
      ;
         H = 't'
      ->
         C = '\t'
      ;
         H = 'b'
      ->
         C = '\b'
      ;
         H = '\''
      ->
         C = '\''
      ;
         H = '\"'
      ->
         C = '\"'
      ;
         H = '\\'
      ->
         C = '\\'
      ;
         H = 'x'
      ->
         'data.char':hexadecimal_digit(Stream:read, High),
         (
            'data.char':hexadecimal_digit(Stream:read, Low)
         ->
            C = 'data.number':char(High * 0x10 + Low)
         ;
            C = 'data.number':char(High)
         )
      ;
         'data.char':decimal_digit(C, D1)
      ->
         (
            'data.char':decimal_digit(Stream:read, D2)
         ->
            (
               'data.char':decimal_digit(Stream:read, D3)
            ->
               C = 'data.number':char((D1 * 10 + D2) * 10 + D3)
            ;
               C = 'data.number':char(D1 * 10 + D2)
            )
         ;
            C = 'data.number':char(D1)
         )
      ).
}
