module 'data.fd' [Impl] {
   /** check(X).
    *  Succeeds if <code>X</code> is a finite-domain variable, the domain
    *  of which is already initialized.
    */

   check(X) :-
      'kernel':fd_check(X).

   /** set_domain(X, Lower, Upper).
    *  <prerequisite>
    *  <code>Lower</code> and <code>Upper</code> are instantiated integers.
    *  </prerequisite>
    *  Succeeds if <code>X</code> is a finite-domain variable, the
    *  domain of which is
    *  <math>[<code>Lower</code> ... <code>Upper</code>]</math>.
    */

   set_domain(X, Lower, Upper) :-
      'kernel':fd_set_domain(X, Lower, Upper).

   /** set_domain(X, List).
    *  <prerequisite>
    *  <code>List</code> is a list of instantiated integers.
    *  </prerequisite>
    *  Succeeds if <code>X</code> is a finite-domain variable, the
    *  domain of which is <code>List</code>.
    */

   set_domain(X, List) :-
      'kernel':fd_set_domain(X, List).

   /** new(Lower, Upper, X).
    *  Equivalent to <code>set_domain(X, Lower, Upper)</code>.
    */

   new(Lower, Upper, X) :-
      set_domain(X, Lower, Upper).

   /** new(List, X).
    *  Equivalent to <code>set_domain(X, List)</code>.
    */

   new(List, X) :-
      set_domain(X, List).

   /** { Domain } = X.
    *  <prerequisite>
    *  Either <code>Domain</code> is a sequence of items of the form <code>Value</code> or <code>Lower .. Upper</code>,
    *  where <code>Value</code>, <code>Lower</code> and <code>Upper</code> are instantiated integers.
    *  </prerequisite>
    *  Succeeds if <code>X</code> is a finite-domain variable, the domain
    *  of which matches the specification given by <code>Domain</code>
    *  (sequence of sparse values and intervals).
    */

   { Domain } = X :-
      (
         Domain = Lower .. Upper
      ->
         new(Lower, Upper, X)
      ;
         new(Impl:domain_list(Domain), X)
      ).

   /** min(X, Lower).
    *  <prerequisite>
    *  <code>X</code> is a finite-domain variable.
    *  </prerequisite>
    *  Succeeds if <code>Lower</code> is the minimal value of the domain of
    *  <code>X</code>.
    */

   min(X) = 'kernel':fd_min(X).

   /** max(X, Upper).
    *  <prerequisite>
    *  <code>X</code> is a finite-domain variable.
    *  </prerequisite>
    *  Succeeds if <code>Upper</code> is the maximal value of the domain of
    *  <code>X</code>.
    */

   max(X) = 'kernel':fd_max(X).

   /** size(X, Size).
    *  <prerequisite>
    *  <code>X</code> is a finite-domain variable.
    *  </prerequisite>
    *  Succeeds if <code>Size</code> is the number of values in the domain of
    *  <code>X</code>.
    */

   size(X) = 'kernel':fd_size(X).

   /** get_domain(X, List).
    *  <prerequisite>
    *  <code>X</code> is a finite-domain variable.
    *  </prerequisite>
    *  Succeeds if <code>List</code> is the list of values in the domain of
    *  <code>X</code>.
    */

   get_domain(X) = 'kernel':fd_dom(X).

   /** (A \/ B) = X.
    *  Succeeds if <code>X</code> is the reified constraint <code>A</code>
    *  &or; <code>B</code>, where <code>A</code> and <code>B</code> are
    *  reified constraints.
    */

   (A \/ B) = 'kernel':fd_or(A, B).

   /** (A /\ B) = X.
    *  Succeeds if <code>X</code> is the reified constraint <code>A</code>
    *  &and; <code>B</code>, where <code>A</code> and <code>B</code> are
    *  reified constraints.
    */

   (A /\ B) = 'kernel':fd_and(A, B).

   /** (A <=> B) = X.
    *  Succeeds if <code>X</code> is the reified constraint <code>A</code>
    *  &hArr; <code>B</code>, where <code>A</code> and <code>B</code> are
    *  reified constraints.
    */

   (A <=> B) = 'kernel':fd_equiv(A, B).

   /** A <=> B.
    *  Ensures that <code>A</code>
    *  &hArr; <code>B</code>, where <code>A</code> and <code>B</code> are
    *  reified constraints.
    */

   A <=> B :- 'kernel':fd_equiv(A, B).

   /** (A => B) = X.
    *  Succeeds if <code>X</code> is the reified constraint <code>A</code>
    *  &rArr; <code>B</code>, where <code>A</code> and <code>B</code> are
    *  reified constraints.
    */

   (A => B) = 'kernel':fd_impl(A, B).

   /** A => B.
    *  Ensures that <code>A</code>
    *  &rArr; <code>B</code>, where <code>A</code> and <code>B</code> are
    *  reified constraints.
    */

   A => B :- 'kernel':fd_impl(A, B).

   /** cardinality(List, X).
    *  <prerequisite>
    *  <code>List</code> is a list of finite-domain variables.
    *  </prerequisite>
    *  Ensures that the finite-domain variable <code>X</code> is equal to
    *  the count of reified constraints which are true in <code>List</code>.
    */

   cardinality(List) = 'kernel':fd_cardinality(List).

   /** all_different(List).
    *  <prerequisite>
    *  <code>List</code> is a list of finite-domain variables.
    *  </prerequisite>
    *  Ensures that the finite-domain variables in <code>List</code> are
    *  pairwise distinct.
    */

   all_different(List) :-
      'kernel':fd_all_different(List)
.
   /** element(I, List, X).
    *  <prerequisite>
    *  <code>List</code> is a list of instantiated integers.
    *  </prerequisite>
    *  Ensures that the <code>I</code>th element of <code>List</code> is
    *  <code>X</code>, where <code>I</code> and <code>X</code> are
    *  finite-domain variables.
    */

   element(I, List, X) :-
      'kernel':fd_element(I, List, X).

   /** element_var(I, List, X).
    *  <prerequisite>
    *  <code>List</code> is a list of finite-domain variables.
    *  </prerequisite>
    *  Ensures that the <code>I</code>th element of <code>List</code> is
    *  <code>X</code>, where <code>I</code> and <code>X</code> are
    *  finite-domain variables.
    */

   element_var(I, List, X) :-
      'kernel':fd_element_var(I, List, X).

   /** labeling(X).
    *  <prerequisite>
    *  <code>X</code> is either a finite-domain variable or a list of
    *  finite-domain variables.
    *  </prerequisite>
    *  Enumerates satisfiable instantiations of <code>X</code>.
    */

   labeling(X) :-
      'kernel':fd_labeling(X).

   /** minimize(M).
    *  <prerequisite>
    *  <code>M</code> is a module with a predicate <code>objective(X)</code>.
    *  </prerequisite>
    *  Finds the minimal value for <code>X</code> by branch-and-bound.
    */

   minimize(M) :-
      'kernel':fd_minimize(M).

   /** maximize(M).
    *  <prerequisite>
    *  <code>M</code> is a module with a predicate <code>objective(X)</code>.
    *  </prerequisite>
    *  Finds the maximal value for <code>X</code> by branch-and-bound.
    */

   maximize(M) :-
      'kernel':fd_maximize(M).

   pretty(X) = 'data.fd:'({ 'data.fd':min(X) .. 'data.fd':max(X) }).
},

module Impl [Impl] {
   domain_list(Number) = [Number] :-
      'data.number':check(Number),
      !.

   domain_list(Lower .. Upper) = 'data.number':(Lower .. Upper).

   domain_list(A, B) = 'data.list':append(Impl:domain_list(A), Impl:domain_list(B)).
}
