module 'data.path' {
   /** path_delimiter(Delim).
    *  Succeeds if <code>Delim</code> is the atom of the path delimiter.
    */

   path_delimiter('/').

   /** concat(A, B, R).
    *  <prerequisite>
    *  <code>A</code> and <code>B</code> are instantiated to paths.
    *  </prerequisite>
    *  If <code>B</code> is absolute, succeeds if <code>R</code> is <code>B</code>.
    *  If <code>B</code> is relative, concats paths <code>A</code> and <code>B</code>,
    *  separated with a path delimiter if needed.
    */

   concat(A, B, R) :-
      'data.path':path_delimiter(Delim),
      (
         'data.path':is_absolute(B)
      ->
         R = B
      ;
         'data.atom':sub(A, _, 1, 0, Delim)
      ->
         'data.atom':concat(A, B, R)
      ;
         'data.atom':(
            concat(A, Delim, A0),
            concat(A0, B, R)
         )
      ).

   /** join(Paths, Path).
    *  <prerequisite>
    *  <code>Paths</code> is a list of instantiated paths.
    *  </prerequisite>
    *  Succeeds if <code>Path</code> is the concatenation of the elements of
    *  <code>Paths</code>, separated with path delimiters.
    */

   join(Paths, Path) :-
      'data.path':path_delimiter(Delim),
      'data.atom':join(Paths, Delim, Path).

   /** split(Path, Base, Filename).
    *  <prerequisite>
    *  <code>Path</code> is an instantiated path.
    *  </prerequisite>
    *  Succeeds if <code>Base</code> is the path of the directory of
    *  <code>Path</code> and <code>Filename</code> is the filename
    *  pointed by <code>Path</code> relative to this directory.
    */

   split(Path, Base, Filename) :-
      'data.path':path_delimiter(Delim),
      (
         'data.atom':split_at_delim_right(Path, Delim, Base, Filename)
      ->
         true
      ;
         Base = '',
         Filename = Path
      ).

   /** base(Path, Base).
    *  Equivalent to <code>split(Path, Base, _)</code>.
    */

   base(Path, Base) :-
      'data.path':split(Path, Base, _).

   /** filename(Path, Filename).
    *  Equivalent to <code>split(Path, _, Filename)</code>.
    */

   filename(Path, Filename) :-
      'data.path':split(Path, _, Filename).

   /** split_suffix(Path, Base, Suffix).
    *  <prerequisite>
    *  <code>Path</code> is an instantiated path.
    *  </prerequisite>
    *  Succeeds if <code>Base</code> is <code>Path</code> with
    *  the suffix removed, and <code>Suffix</code> is this suffix,
    *  dot excepted.
    */

   split_suffix(Path, Base, Suffix) :-
      'data.path':path_delimiter(Delim),
      module Aux [Delim, Path, Base, Suffix] {
         split(Index) :-
            'data.number':(Index >= 0),
            'data.atom':sub(Path, Index, 1, _, C),
            (
               C = '.'
            ->
               'data.atom':sub(Path, 0, Index, _, Base),
               'data.atom':sub(Path, 'data.number':succ(Index), _, 0, Suffix)
            ;
               C \= Delim,
               this:split('data.number':pred(Index))
            ). 
      },
      (
         Aux:split('data.number':pred('data.atom':length(Path)))
      ->
         true
      ;
         Base = Path,
         Suffix = ''
      ).

   /** suffix(Path, Suffix).
    *  Equivalent to <code>split_suffix(Path, _, Suffix)</code>.
    */

   suffix(Path, Suffix) :-
      'data.path':split_suffix(Path, _, Suffix).

   /** chop_suffix(Path, Base).
    *  Equivalent to <code>split_suffix(Path, Base, _)</code>.
    */

   chop_suffix(Path, Base) :-
      'data.path':split_suffix(Path, Base, _).

   /** is_absolute(Path).
    *  Succeeds of <code>Path</code> is an absolute path.
    */

   is_absolute(Path) :-
      'data.path':path_delimiter(Delim),
      'data.atom':sub(Path, 0, 1, _, Delim).
}
